void __svgalib_driver8_setpixel(int, int, int);
int __svgalib_driver8_getpixel(int, int);
void __svgalib_driver8_hline(int, int, int, int);
void __svgalib_driver8_fillbox(int, int, int, int, int);
void __svgalib_driver8_putbox(int, int, int, int, void *, int);
void __svgalib_driver8_getbox(int, int, int, int, void *, int);
void __svgalib_driver8_putboxmask(int, int, int, int, void *);
void __svgalib_driver8_putboxpart(int, int, int, int, int, int, void *, int, int);
void __svgalib_driver8_getboxpart(int, int, int, int, int, int, void *, int, int);
void __svgalib_driver8_copybox(int, int, int, int, int, int);

void __svgalib_driver16_setpixel(int, int, int);
int __svgalib_driver16_getpixel(int, int);
void __svgalib_driver16_hline(int, int, int, int);
void __svgalib_driver16_fillbox(int, int, int, int, int);
void __svgalib_driver16_putbox(int, int, int, int, void *, int);
void __svgalib_driver16_getbox(int, int, int, int, void *, int);
void __svgalib_driver16_putboxmask(int, int, int, int, void *);
void __svgalib_driver16_putboxpart(int, int, int, int, int, int, void *, int, int);
void __svgalib_driver16_getboxpart(int, int, int, int, int, int, void *, int, int);
void __svgalib_driver16_copybox(int, int, int, int, int, int);

void __svgalib_driver24_setpixel(int, int, int);
int __svgalib_driver24_getpixel(int, int);
void __svgalib_driver24_hline(int, int, int, int);
void __svgalib_driver24_fillbox(int, int, int, int, int);
void __svgalib_driver24_putbox(int, int, int, int, void *, int);
void __svgalib_driver24_getbox(int, int, int, int, void *, int);
void __svgalib_driver24_putboxmask(int, int, int, int, void *);
void __svgalib_driver24_putboxpart(int, int, int, int, int, int, void *, int, int);
void __svgalib_driver24_getboxpart(int, int, int, int, int, int, void *, int, int);
void __svgalib_driver24_copybox(int, int, int, int, int, int);
void __svgalib_driver24_putbox32(int, int, int, int, void *, int);

void __svgalib_driver32_setpixel(int, int, int);
int __svgalib_driver32_getpixel(int, int);
void __svgalib_driver32_hline(int, int, int, int);
void __svgalib_driver32_fillbox(int, int, int, int, int);
void __svgalib_driver32_putbox(int, int, int, int, void *, int);
void __svgalib_driver32_getbox(int, int, int, int, void *, int);
void __svgalib_driver32_putboxmask(int, int, int, int, void *);
void __svgalib_driver32_putboxpart(int, int, int, int, int, int, void *, int, int);
void __svgalib_driver32_getboxpart(int, int, int, int, int, int, void *, int, int);
void __svgalib_driver32_copybox(int, int, int, int, int, int);

void __svgalib_driver8p_setpixel(int, int, int);
int __svgalib_driver8p_getpixel(int, int);
void __svgalib_driver8p_hline(int, int, int, int);
void __svgalib_driver8p_fillbox(int, int, int, int, int);
void __svgalib_driver8p_putbox(int, int, int, int, void *, int);
void __svgalib_driver8p_getbox(int, int, int, int, void *, int);
void __svgalib_driver8p_putboxmask(int, int, int, int, void *);
void __svgalib_driver8p_putboxpart(int, int, int, int, int, int, void *, int, int);
void __svgalib_driver8p_getboxpart(int, int, int, int, int, int, void *, int, int);
void __svgalib_driver8p_copybox(int, int, int, int, int, int);

void __svgalib_driver16p_setpixel(int, int, int);
int __svgalib_driver16p_getpixel(int, int);
void __svgalib_driver16p_hline(int, int, int, int);
void __svgalib_driver16p_fillbox(int, int, int, int, int);
void __svgalib_driver16p_putbox(int, int, int, int, void *, int);
void __svgalib_driver16p_getbox(int, int, int, int, void *, int);
void __svgalib_driver16p_putboxmask(int, int, int, int, void *);
void __svgalib_driver16p_putboxpart(int, int, int, int, int, int, void *, int, int);
void __svgalib_driver16p_getboxpart(int, int, int, int, int, int, void *, int, int);
void __svgalib_driver16p_copybox(int, int, int, int, int, int);

void __svgalib_driver24p_setpixel(int, int, int);
int __svgalib_driver24p_getpixel(int, int);
void __svgalib_driver24p_hline(int, int, int, int);
void __svgalib_driver24p_fillbox(int, int, int, int, int);
void __svgalib_driver24p_putbox(int, int, int, int, void *, int);
void __svgalib_driver24p_getbox(int, int, int, int, void *, int);
void __svgalib_driver24p_putboxmask(int, int, int, int, void *);
void __svgalib_driver24p_putboxpart(int, int, int, int, int, int, void *, int, int);
void __svgalib_driver24p_getboxpart(int, int, int, int, int, int, void *, int, int);
void __svgalib_driver24p_copybox(int, int, int, int, int, int);

void __svgalib_driver32p_setpixel(int, int, int);
int __svgalib_driver32p_getpixel(int, int);
void __svgalib_driver32p_hline(int, int, int, int);
void __svgalib_driver32p_fillbox(int, int, int, int, int);
void __svgalib_driver32p_putbox(int, int, int, int, void *, int);
void __svgalib_driver32p_getbox(int, int, int, int, void *, int);
void __svgalib_driver32p_putboxmask(int, int, int, int, void *);
void __svgalib_driver32p_putboxpart(int, int, int, int, int, int, void *, int, int);
void __svgalib_driver32p_getboxpart(int, int, int, int, int, int, void *, int, int);
void __svgalib_driver32p_copybox(int, int, int, int, int, int);

void __svgalib_driver8a_fillbox(int, int, int, int, int);
void __svgalib_driver8a_copybox(int, int, int, int, int, int);

void __svgalib_driverplanar256_nothing(void);
void __svgalib_driverplanar256_putbox(int, int, int, int, void *, int);

void __svgalib_driverplanar16_nothing(void);

/* Generic functions */

int __svgalib_driver_setread(GraphicsContext * gc, int i, void **vp);
int __svgalib_driver_setwrite(GraphicsContext * gc, int i, void **vp);

/* internal globals: */
extern void (*__svgalib_nonaccel_fillbox)(int, int, int, int, int);
