
extern void __svgalib_set_graphtermio(void);
extern void __svgalib_enable_interrupt(void);
extern void __svgalib_disable_interrupt(void);
extern void __svgalib_waitvtactive(void);
extern void __svgalib_open_devconsole(void);
extern void __svgalib_set_texttermio(void);


